/*  ZA2 Driver for Linux
    Copyright (C) 1998,1999,2000 Peter Wahl

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Send bug reports to:
    Peter Wahl
    uzsa69@uni-bonn.de
*/

#define __NO_VERSION__
#include "za2.h"
#ifndef NO_UACCESS
#include <asm/uaccess.h>
#endif
#include "sound_config.h"
#include "za2hard.h"
#include "za2recplay.h"
#include "za2mixer.h"
#include "za2oldkernels.h"

/***  za2_read ***/ 
#ifndef OLD_FILEOPS
static ssize_t za2_read(struct file *file, char * buf, size_t count, loff_t *lp)
#else
static ssize_t za2_read(struct inode *inode,struct file *file, char * buf, int count)
#endif
{
  za2_status_t *dev=za2_st;
  void *ptr;
  int actualCount = 0,c=count;
#ifdef NO_SOUND_H
#ifndef OLD_FILEOPS
  if(MINOR(file->f_dentry->d_inode->i_rdev)==dev->minor)
#else
  if(MINOR(inode->i_rdev)==dev->minor)
#endif
#endif
    while(c)
      {
	if( (actualCount =  za2_begin_read(dev,&ptr, (size_t) c)) < 0)   /* get a pointer on dma buffer (this can block) */
	  return actualCount;
	copy_to_user(buf, ptr, actualCount);                                  /* copy actual data */ 
	c -= actualCount;
	buf += actualCount;
      }
  return count;
}

/***  za2_write ***/
#ifndef OLD_FILEOPS
static ssize_t za2_write(struct file *file, const char * buf, size_t count, loff_t *lp) 
#else
static ssize_t za2_write(struct inode *inode,struct file *file, const char * buf, int count) 
#endif
{
  za2_status_t *dev=za2_st;
  void *ptr;
  int actualCount = 0,c=count;
#ifdef NO_SOUND_H
#ifndef OLD_FILEOPS
  if(MINOR(file->f_dentry->d_inode->i_rdev)==dev->minor)
#else
  if(MINOR(inode->i_rdev)==dev->minor)
#endif
#endif
    while(c)
      {
	switch(dev->filter)
	  {
	  case MPEG2:
	    if( (actualCount = za2_write_mp2(dev,&ptr, (size_t) c)) < 0)  /* get a pointer on dma buffer (this can block) */
	      return actualCount;
	    break;
	  default:
	    if( (actualCount = za2_begin_write(dev,&ptr, (size_t) c)) < 0)  /* get a pointer on dma buffer (this can block) */
	      return actualCount;
	  }
	copy_from_user(ptr, buf, actualCount);                    	      /* copy actual data */
	c -= actualCount;
	buf += actualCount;
      }
  return count;
}

/*Handles IOCTL Requests*/
static int za2_ioctl(struct inode * aInode, struct file * aFile,
                     unsigned int cmd, unsigned long arg)
{
  int  err   = 0,inputrate,buf;
  long param = 0;
  za2_status_t *dev=za2_st;
  za2_dma_t *dmap=dev->outdma;
  audio_buf_info *abi;
  count_info *cnti;
#ifdef NO_SOUND_H
  if(MINOR(aInode->i_rdev)==dev->mixer)
    return mixer_ioctl(aInode,aFile,cmd,arg);
#endif
  if(((cmd>>8)&0xff) == 'P') 
    {
      /* extract the parameter */
      if((_SIOC_DIR(cmd)&_SIOC_WRITE) && (_SIOC_SIZE(cmd)<=sizeof(long))) 
	{
	  err = verify_area(VERIFY_READ, (void *) arg, sizeof(long));
	  if(err)
	    return err;
#ifndef OLD_GET_USER
	  get_user(param,(long *) arg);
#else
	  param=get_user((long *) arg);
#endif
	}
      /* do the command */
      switch(cmd) 
	{
	case SNDCTL_DSP_RESET: 
	case SNDCTL_DSP_SYNC: 
	case SNDCTL_DSP_POST: 
	  break;
	case SNDCTL_DSP_SPEED: 
	  /* in play mode (i.e. not rec nor recplay with record enabled), */ 
	  /* we always change the rate */
	  if(!(dev->currentState&ZA2_OPEN_REC))
	    {
	      err = zaSetSamplingRate(dev,param);
	      param = zaGetSamplingRate(dev);
	    }
	  /* in record mode, we check for matching rates                       */
	  /* If the rates don't match, we try to load filters (not in recplay) */
	  else if(dev->currentState==ZA2_OPEN_REC) 
	    { 
	      if((inputrate=zaGetSamplingRate(dev)) == param)
#ifdef DSPOSFX_SUPPORT
		err=loadfilter(dev,dev->dspcode);
#else
		err=loadfilter(dev,DSPOS);
#endif
	      else if((inputrate==32000)&&(param==44100)) err=loadfilter(dev,RESAMPLE_32244);
	      else if((inputrate==48000)&&(param==44100)) err=loadfilter(dev,RESAMPLE_48244);
	      else { param=inputrate; err=1; }        /*else return current frequency*/
	      if((err==0) && (dev->currentState&ZA2_OPEN_REC))
		{
		  dev->currentState=ZA2_CLOSE;
		  za2_open_rec(dev);
		}
	    }
	  /*  in recplay, we only return the sampling rate of the input,  */ 
	  /*  if there is any  */
	  else param = zaGetSamplingRate(dev); 
	  if(err<0) /* unsupported rate or error */
	    return err;
	  break;
	case SNDCTL_DSP_STEREO: 
	  if(param != 1)
	    {
	      printk("ZA2: ioctl(SNDCTL_DSP_STEREO,%ld) not supported.\n",param);
	      return ZA2_FEATURE_NOT_SUPPORTED;
	    }
	  break;
	case SNDCTL_DSP_GETBLKSIZE: 
	  param=za2_get_block_size(dev);
	  break;
	case SNDCTL_DSP_GETCAPS:
	  param=DSP_CAP_DUPLEX;
	  break;
	case SNDCTL_DSP_SETFMT:
	  switch(param)
	    {
	    case AFMT_MPEG : 
	      if(dev->currentState==ZA2_OPEN_PLAY) 
		loadfilter(dev,MPEG2);
	      else return ZA2_FEATURE_NOT_SUPPORTED;
	      break;
	    case AFMT_S16_LE :
	      if(!(dev->filter&7))
#ifdef DSPOSFX_SUPPORT 
		loadfilter(dev,dev->dspcode);                 /* not resampler or dspos */
#else
	        loadfilter(dev,DSPOS);                 /* not resampler or dspos */
#endif
	      break;
	    case AFMT_QUERY :
	      if(dev->filter&7)
		param=AFMT_S16_LE;
	      else param=AFMT_MPEG;
	      break;
	    default :
	      printk("ZA2: ioctl(SNDCTL_DSP_SETFMT,%lx) not supported.\n",param);
	      return ZA2_FEATURE_NOT_SUPPORTED;
	    }
	  break;
	case SNDCTL_DSP_CHANNELS: 
	  if(param == 1)
	    {
	      printk("ZA2: ioctl(SNDCTL_DSP_CHANNELS,%ld) not supported.\n",param);
	      return ZA2_FEATURE_NOT_SUPPORTED;
	    }
	  param = 2;
	  break;
	case SNDCTL_DSP_SUBDIVIDE: 
	  if((param==1) || (param==2) || (param==4))
	    dev->blocksize=ZA2_BLOCK_SIZE/param;
	  else return -EINVAL;
	  break;
	case SNDCTL_DSP_SETFRAGMENT:
	  buf=1<<param;
	  if((buf<((MAX_BUFFER+1)>>1)) && (buf>128)) 
	    dev->blocksize=buf;
	  else return -EINVAL;
	  break;
	case SNDCTL_DSP_GETFMTS: 
	  param = AFMT_S16_LE|AFMT_MPEG;
	  break;
	case SNDCTL_DSP_GETISPACE:
	  dmap=dev->indma;
	case SNDCTL_DSP_GETOSPACE:
	  abi=(audio_buf_info *)vmalloc(sizeof(audio_buf_info));
	  if(abi)
	    {
	      if(!dmap->firstBlock)
		{
		  buf=dmap->currentDmaPtr-dmap->currentDataPtr;
		  if(buf<0) buf+=MAX_BUFFER;
		}
	      else buf=MAX_BUFFER;
	      abi->fragments=buf/dev->blocksize;
	      abi->fragstotal=MAX_BUFFER/dev->blocksize;
	      abi->fragsize=dev->blocksize;
	      abi->bytes=buf;
	      copy_to_user((void *) arg,abi,sizeof(audio_buf_info));
	      vfree(abi);
	    }
	  else return -ENOMEM;
	  break;
	case SNDCTL_DSP_GETIPTR:
	  if(dev->currentState!=ZA2_OPEN_REC) return -EINVAL;
	  dmap=dev->indma;
	case SNDCTL_DSP_GETOPTR:
	  if((cmd==SNDCTL_DSP_GETOPTR) && (dev->currentState!=ZA2_OPEN_PLAY)) return -EINVAL;
	  cnti=(count_info *)vmalloc(sizeof(count_info));
	  if(cnti)
	    {
	      cnti->bytes=dev->interrupt*dev->blocksize;
	      cnti->blocks=dev->interrupt;
	      cnti->ptr=dmap->currentDmaPtr;
	      copy_to_user((void *) arg,cnti,sizeof(count_info));
	      vfree(cnti);
	    }
	  else return -ENOMEM;
	  break;
	default:
	  printk("ZA2: tried to set %x.\n",cmd);
	  return -1;
	}
      /* put back the parameter */
      if((_SIOC_DIR(cmd)&_SIOC_READ)&&(_SIOC_SIZE(cmd)<=sizeof(long))) 
	{
	  err = verify_area(VERIFY_WRITE, (void *) arg, sizeof(long));
	  if(err)
	    return err;
	  put_user(param, (long *) arg);
	}
      return 0;  /* all other settings are correct, but not handled */
    }
  else return mixer_ioctl(aInode,aFile,cmd,arg);
}

/***  za2_open ***/
static int za2_open(struct inode * aInode, struct file * aFile)
{
  int err;
  za2_status_t *dev=za2_st;
#ifdef NO_SOUND_H
  if(MINOR(aInode->i_rdev)==dev->mixer)
    {
#ifdef USE_MODCOUNT
      MOD_INC_USE_COUNT;
#endif
      return 0;
    }
#endif
  if(dev->currentState)
    {
      DEBUG(printk("ZA2: Device busy.\n"));
      return -EBUSY;
    }
  if(!dev->initialized) 
    {
      DEBUG(printk("ZA2: Not initialized.\n"));
      return -1;
    }
  if( aFile->f_flags & O_RDWR) 
    {
#ifdef DSPOSFX_SUPPORT
      loadfilter(dev,dev->dspcode);
#else
      loadfilter(dev,DSPOS);
#endif
      dev->currentState=ZA2_OPEN_RECPLAY;
      za2_open_recplay(dev); 
      DEBUG(printk("ZA2: open for playing/recording, 16 bit\n"));
    }
  else if( (aFile->f_flags & O_ACCMODE) == O_RDONLY ) 
    {
#ifdef DSPOSFX_SUPPORT
      loadfilter(dev,dev->dspcode);
#else
      loadfilter(dev,DSPOS);
#endif
      err = za2_open_rec(dev);
      DEBUG(printk("ZA2: open for recording, 16 bit\n"));
      if(err) 
	{
	  DEBUG(printk("ZA2: open rec: err = %d\n", err));
	  return -1;
	}
    }
  else if( (aFile->f_flags & O_ACCMODE) == O_WRONLY) 
    {
#ifdef DSPOSFX_SUPPORT
      loadfilter(dev,dev->dspcode);
#else
      loadfilter(dev,DSPOS);
#endif
      err = za2_open_play(dev);
      DEBUG(printk("ZA2: open for playing, 16 bit\n"));
      if(err) 
	{
	  DEBUG(printk("ZA2: open play: err = %d\n", err));
	  return -1;
	}
    }
  else 
    {
      DEBUG(printk("ZA2: Unknown open flag\n"));
      return -1;
    }
#ifdef USE_MODCOUNT
  MOD_INC_USE_COUNT;
#endif
  return 0;
}

/***  za2_release ***/
#ifndef OLD_FILEOPS
static int za2_release(struct inode * aInode, struct file * aFile)
#else
static void za2_release(struct inode * aInode, struct file * aFile)
#endif
{
  za2_status_t *dev=za2_st;
#ifdef NO_SOUND_H
  if(MINOR(aInode->i_rdev)==dev->mixer)
    {
#ifdef USE_MODCOUNT
      MOD_DEC_USE_COUNT;
#endif
      return 0;
    }
#endif
  za2_close(dev);
#ifdef USE_MODCOUNT
  MOD_DEC_USE_COUNT;
#endif
#ifndef OLD_FILEOPS
  return(0);
#endif
}

static struct file_operations za2card_fops = {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
  THIS_MODULE,
#endif
  NULL,
  za2_read,
  za2_write,
  NULL,
  NULL,
  za2_ioctl,
  NULL,
  za2_open,
#ifndef OLD_FILEOPS
  NULL,
#endif
  za2_release,
  NULL,
  NULL,
  NULL,
  NULL
#ifndef OLD_FILEOPS
 ,NULL
#endif
};
