/*  ZA2 Driver for Linux
    Copyright (C) 1998,1999,2000 Peter Wahl

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Send bug reports to:
    Peter Wahl
    uzsa69@uni-bonn.de
*/

#define __NO_VERSION__
#include "za2.h"
#include <asm/io.h>
#include <linux/soundcard.h>
#ifdef USE_DELAY
#include <asm/delay.h>
#endif
#include <linux/delay.h>

#ifdef NO_MDELAY
void __inline__ mdelay(int val)
{
  int i;
  for(i=0;i<val;i++)
    udelay(1000);
}
#endif

#define _za2hard_c
#ifndef NO_SOUND_H
#include "sound_firmware.h"
#else
#include "linux/sound_firmware.h"
#endif
#include "za2hard.h"

unsigned short __inline__ za2status(za2_status_t *dev)
{
  return((unsigned short)inw(dev->base+2));
}

void __inline__ za2command(za2_status_t *dev,short val)
{
  outw(val,dev->base+2);
}

void __inline__ disable_za2(za2_status_t *dev)
{
  za2command(dev,dev->command&0xff03);
}

void __inline__ mute_za2(za2_status_t *dev)
{
  outw(0,dev->base);
}

int __inline__ za2setcommand(za2_status_t *dev)
{
  return(dev->command&(0xff03|((dev->recording)?0x3c:0)|((dev->playing)?0xcc:0)));
}

void __inline__  za2setmode(za2_status_t *dev)
{
  za2command(dev,za2setcommand(dev));
}

short wait_za2(za2_status_t *dev)
{
  int timeout=0;
#ifdef USE_DELAY
  while((za2status(dev)&8) && (timeout++<20)) udelay(1);
  while(!(za2status(dev)&8) && (timeout++<40)) udelay(1);
  return(timeout==40);
#else
  while((za2status(dev)&8) && (timeout++<40));
  while(!(za2status(dev)&8) && (timeout++<80));
  return(timeout==80);
#endif
}

void send_byte(za2_status_t *dev,short val)
{
  int timeout=0;
#ifdef USE_DELAY
  while((za2status(dev)&0x400) && (timeout++<80)) udelay(1);
  if(timeout<80)
    outw(val, dev->base+4);
#else
  while((za2status(dev)&0x400) && (timeout++<160));
  if(timeout<160)
    outw(val, dev->base+4);
#endif
  DEBUG(else printk("ZA2: send_byte : Timeout !\n"));
}

void ZaSetEnv(za2_status_t *dev)
{
  dev->command = 0x1000 + (dev->mode << 8) + dev->digin + ((dev->irq - 9) << 2);
  if(dev->indma)
    dev->command+=((dev->indma->dma&3)<<4);
  if(dev->outdma)
    dev->command+=((dev->outdma->dma&3)<<6);
}

/*  zaGetSamplingRate */
unsigned int zaGetSamplingRate(za2_status_t *dev)
{
  int srate=0;
  if(dev->mode&2)                                  /* Check only in record and sync-mode*/
    {
      switch(za2status(dev)&0x830)
	{
	case 0x000:
          DEBUG(printk("Out of range...\n"));
	  srate = 48000;
	  break;
	case 0x010:
          DEBUG(printk("48Khz (rough)\n"));
	  srate = 48000;
	  break;
	case 0x020:
	  DEBUG(printk("44.1Khz (rough)\n"));
	  srate = 44100;
	  break;
	case 0x030:
	  DEBUG(printk("32Khz (rough)\n"));
	  srate = 32000;
	  break;
	case 0x800:
	  DEBUG(printk("48Khz\n"));
	  srate = 48000;
	  break;
	case 0x810:
	  DEBUG(printk("44.1Khz\n"));
	  srate = 44100;
	  break;
	case 0x820:
	  DEBUG(printk("44.056Khz\n"));
	  srate = 44056;
	  break;
	case 0x830:
	  DEBUG(printk("32Khz\n"));
	  srate = 32000;
	  break;
	}
      if((dev->filter&7) && ((dev->filter&7)!=DSPOS)) 
	{
	  DEBUG(printk("resampled"));
	  srate = 44100;
	}
      if(dev->filter==MPEG2) {
	DEBUG(printk("MPEG2"));
      }
    }
  else { 
    srate=dev->srate; 
    DEBUG(printk("%d Hz\n",srate));
  }
  // dev->srate=srate;
  return srate;
}

/*** ZaOpenDSP ***/
static void ZaOpenDSP(za2_status_t *dev)
{
  int actcom;
  actcom=za2setcommand(dev);
  za2command(dev,0x3000 + (actcom&0x2ff));         /* raise CS */
  wait_za2(dev);
  outw(0x0, dev->base+4);                          /* clock it.. 8 times actually */
  wait_za2(dev);
  za2command(dev,0x1000 + (actcom&0x2ff));         /* drop CS again, keep reset hi */
  wait_za2(dev);
  outw(0, dev->base+4);                            /* send out read bit and dummy address */
  wait_za2(dev);
  outw(0, dev->base+4);
  wait_za2(dev);
  outw(0, dev->base+4);
  wait_za2(dev);
  outw(0, dev->base+4);
}

/*** SendDSPCMD ***/
static void ZaSendDSPCMD(za2_status_t *dev,unsigned short cmd, unsigned int data)
{
  /*wait_za2(dev);
  outw(0, dev->base+4);
  wait_za2(dev);
  outw(0, dev->base+4);
  wait_za2(dev);
  outw(0, dev->base+4); *TEST*/
  wait_za2(dev);
  outw(cmd, dev->base+4);
  cmd = ((data&0xff0000)>>16);
  wait_za2(dev);
  outw(cmd, dev->base+4);
  cmd = ((data&0xff00)>>8);
  wait_za2(dev);
  outw(cmd, dev->base+4);
  cmd = (data&0xff);
  wait_za2(dev);
  outw(cmd, dev->base+4);
}

/*** ZaCloseDSP ** (Return!=0 -> Error)*/
static short ZaCloseDSP(za2_status_t *dev)
{
  int wait,w2,actcom;
  short last,samp,ercflag=0,r;
  actcom=za2setcommand(dev);
  za2command(dev,0x1000 + (actcom&0x2ff));         /* be sure we have a clock */
  ercflag += wait_za2(dev); 
  outw(0x00ff, dev->base+4);                       /* signal end of block...? */
  ercflag += wait_za2(dev);
  outw(0x00ff, dev->base+4);                       /* signal end of block...? */
  ercflag += wait_za2(dev);
  za2command(dev,0x3000 + (actcom&0x2ff));         /* raise CS */
  ercflag += wait_za2(dev);
  outw(0x0, dev->base+4);                          /* clock it.. 8 times actually */
  ercflag += wait_za2(dev);
  za2command(dev,0x1000 + (actcom&0x2ff));         /* drop CS again, keep reset hi */
  ercflag += wait_za2(dev);
  outw(0x1, dev->base+4);                          /* send out read bit and dummy address */
  ercflag += wait_za2(dev);
  za2command(dev,0x1000 + ((actcom&0x2ff)|0x100)); /*  let the PC read info */
  if (ercflag) return(7);                          /* 7 means Clock is dead.. */
  wait = 0;
  while(!(za2status(dev)&0x200))                   /* make sure REQ is inactive */
    {
      if(wait++>1000)
	return(8);                                 /* 8 is 'REQ never went inactive' */
      w2=MAX_LOOP;
      while((za2status(dev)&0x400) && --w2)        /* wait for host port */
#ifdef USE_DELAY 
	udelay(1) 
#endif
	  ;
      if(!w2) 
	{
	  DEBUG(printk("ZA2:CloseDSP: Timeout !!\n"));
	  return -1;
	}
      outw(0, dev->base+4);                        /* clock it 8 times */
    }
  za2setmode(dev);                                 /* back to normal */
  wait = 0;
  while(za2status(dev)&0x200)                      /* wait for REQ to be active */
    {
      if(wait++>2000)                              /* could be a long wait.. not sure why */
	return(9);                                 /* 9 is 'REQ never came back!' */
      wait_za2(dev);
    }
  if (dev->mode&1)                                 /* wait for the 'za' sync for mode 1 and 3 */
    {
      wait = 0;
      samp = 0;
      last=(za2status(dev)&8); 
      while(samp!=0x7a61)
	{
	  w2=MAX_LOOP;
	  while(((za2status(dev)&8)==last) && --w2) 
#ifdef USE_DELAY
	    udelay(1)
#endif
	      ;
	  if(!w2) 
	    {
	      DEBUG(printk("ZA2:CloseDSP: Timeout !!\n"));
	      return -1;
	    }
	  last = (za2status(dev)&8);
	  samp = (short) inw(dev->base);
	  if(wait++>100)
	    {
	      if (za2status(dev)&0x200)
		return(3);
	      else
		return(1);
	    }
	}
      for(r=0;r<16;r++)
	{
	  w2=MAX_LOOP;
	  while(((za2status(dev)&8)==last) && --w2) 
#ifdef USE_DELAY
	    udelay(1)
#endif
	      ;
	  if(!w2) 
	    {
	      DEBUG(printk("ZA2:CloseDSP: Timeout !!\n"));
	      return -1;
	    }
	  last = (za2status(dev)&8);
	  if( ((unsigned short) inw(dev->base)) != 0x7a61)
	    {
	      if (za2status(dev)&0x200)
		return(4);
	      else
		return(2);
	    }
	}
    }
  return(0);
}

/*** ZaNuMode ***/
static short ZaNuMode(za2_status_t *dev,short numode)
{
  int stillLoop = MAX_LOOP,err=0;
  if((dev->mode&2)!=(numode&2) || (!dev->mode))
  {
    za2command(dev,0x1000+((1+(numode&2)) << 8)+dev->digin);
    wait_za2(dev);
    wait_za2(dev);
    while((za2status(dev)&0x200)&&(--stillLoop)) wait_za2(dev);
    if(!stillLoop) {
	DEBUG(printk("ZA2: Got no sync !\n"));
	err=-EAGAIN;
    }
  }
  za2command(dev,0x1000+(numode << 8)+dev->digin);
  dev->mode = numode;
  dev->command &= 0xfcff;
  dev->command += (dev->mode << 8);
  return err;
}

/* Ldt DSP-Code auf Karte */
short up_sim(za2_status_t *dev,char *code,int len,int program)
{
  unsigned char *za2_ptr=code;
  unsigned short r,adrs,length;
  unsigned long flags;
  long tcheck,checksum = 0;
  int pp = 0;
  ZaSetEnv(dev);
  za2command(dev,0x7000);     /* raise BOOT */
  wait_za2(dev);
  za2command(dev,0x5000);     /* select SPI mode, drop CS */
  wait_za2(dev);
  za2command(dev,0x4000);     /* drop RESET */
  wait_za2(dev);
  za2command(dev,0x5000);     /* raise RESET, initiate download */
  wait_za2(dev); 
  za2command(dev,0x1000);     /* drop BOOT */
  wait_za2(dev); 
  delay(100);                 /* pause */
  /* begin of upload process */
  za2command(dev,0x3000);     /* raise CS */
  wait_za2(dev);
  outw(0, dev->base+4);       /* clock it.. 8 times actually */
  wait_za2(dev);
  za2command(dev,0x1000);     /* drop CS again, keep reset hi */
  wait_za2(dev);
  outw(0, dev->base+4);       /* send out write bit and dummy address */
  wait_za2(dev);
  GET_CPU_EXCLUSIVITY;
  while(pp<len)
  {
    send_byte(dev,*za2_ptr);
    adrs=((*za2_ptr) << 8);
    za2_ptr++; pp++;
    send_byte(dev,*za2_ptr);
    adrs+=*za2_ptr;
    za2_ptr++; pp++;
    if(adrs==0xffff)
    {
      send_byte(dev,*(za2_ptr++)); pp++;
      send_byte(dev,*(za2_ptr++)); pp++;
      send_byte(dev,*(za2_ptr++)); pp++;
      wait_za2(dev);
      if(!(za2status(dev)&0x200))
	{
	  disable_za2(dev);	  
	  RELEASE_CPU_EXCLUSIVITY;
	  printk("ZA2: Upload of DSP-OS failed.\n");
	  return ZA2_DOWNLOAD_FAILED;
	}
      switch(program)
	{
	case RESAMPLE_48244 :
	  DEBUG(printk("48Khz"));
	  ZaNuMode(dev,3);
	  dev->filter=RESAMPLE_48244;
	  dev->blocksize=ZA2_BLOCK_SIZE;
#ifdef DSPOSFX_SUPPORT
	  dev->mixer_devs=SOUND_MASK_VOLUME;
	  dev->mixer_caps=0;
#endif
	  DEBUG(printk(" to 44.1Khz resampler loaded\n"));
	  break;
	case RESAMPLE_32244 :
	  DEBUG(printk("32Khz"));
	  ZaNuMode(dev,3);
	  dev->filter=RESAMPLE_32244;
	  dev->blocksize=ZA2_BLOCK_SIZE;
#ifdef DSPOSFX_SUPPORT
	  dev->mixer_devs=SOUND_MASK_VOLUME;
	  dev->mixer_caps=0;
#endif
	  DEBUG(printk(" to 44.1Khz resampler loaded\n"));
	  break;
	case DSPOS :
	  DEBUG(printk("DSP"));
	  dev->mode=0;
	  if(dev->currentState != ZA2_OPEN_REC)   /*to be tested !!*/
	    ZaNuMode(dev,1);
	  dev->filter=DSPOS;
	  dev->blocksize=ZA2_BLOCK_SIZE;
	  dev->lavol = 16384;
#ifdef DSPOSFX_SUPPORT
	  dev->mixer_devs=SOUND_MASK_VOLUME;
	  dev->mixer_caps=0;
#endif
	  DEBUG(printk(" Operating System loaded\n"));
	  break;
#ifdef DSPOSFX_SUPPORT
	case DSPOSFX :
	  DEBUG(printk("DSPFX"));
	  dev->mode=0;
	  if(dev->currentState != ZA2_OPEN_REC)   /*to be tested !!*/
	    ZaNuMode(dev,1);
	  dev->filter=DSPOSFX;
	  dev->blocksize=ZA2_BLOCK_SIZE;
	  dev->mixer_devs=SOUND_MASK_VOLUME|SOUND_MASK_LINE;
	  dev->mixer_caps=SOUND_MASK_VOLUME|SOUND_MASK_LINE;
	  dev->ravol=dev->lavol=dev->rdvol=dev->ldvol=16384;
	  DEBUG(printk(" Operating System loaded\n"));
	  break;
#endif
	case MPEG2 :
	  DEBUG(printk("MPEG2"));
	  ZaNuMode(dev,1);
	  dev->filter=MPEG2;
	  dev->blocksize=MP2BLOCKSIZE;
#ifdef DSPOSFX_SUPPORT
	  dev->mixer_devs=0;
	  dev->mixer_caps=0;
#endif
	  DEBUG(printk(" Decoder loaded\n"));
	  break;
	default :
	  DEBUG(printk("User program"));
	  ZaNuMode(dev,1);
	  dev->filter=USER;
#ifdef DSPOSFX_SUPPORT
	  dev->mixer_devs=0;
	  dev->mixer_caps=0;
#endif
	  DEBUG(printk(" %d loaded (Mode 1)",program));
	}
      dev->srate = 48000;
      disable_za2(dev);
      mute_za2(dev);
      RELEASE_CPU_EXCLUSIVITY;
      return 0;
    }
    checksum+=(long)adrs;
    send_byte(dev,*za2_ptr);
    length=((*za2_ptr) << 8);
    za2_ptr++; pp++;
    send_byte(dev,*za2_ptr);
    length+=*za2_ptr;
    za2_ptr++; pp++;
    checksum+=(long)length;
    length= length/3;
    for (r=0;r<length;r++)
    {
      send_byte(dev,*za2_ptr);
      tcheck=((long)(*za2_ptr) << 16);
      za2_ptr++; pp++;
      send_byte(dev,*za2_ptr);
      tcheck+=((long)(*za2_ptr) << 8);
      za2_ptr++; pp++;
      send_byte(dev,*za2_ptr);     
      tcheck+=(long)(*za2_ptr);
      za2_ptr++; pp++;
      checksum += tcheck;
    }
  }
  RELEASE_CPU_EXCLUSIVITY;
  return ZA2_UPSIM_UNKNOW_ERROR;
}

/* Return: >0: nothing happened, 0: loaded successfully, <0: error */
int loadfilter(za2_status_t *dev,int nr)
{
  char *firmware;
  int len;
  if(dev->filter!=nr)
    {
      DEBUG(printk("Loading Firmware..."));
      switch(nr)
	{
	case DSPOS : 
	  DEBUG(printk("DSPOS "));
	  len = mod_firmware_load(ZA2FIRMWAREDIR "/dsposx2.sim", (char **) &firmware);
	  break;
#ifdef DSPOSFX_SUPPORT
	case DSPOSFX : 
	  DEBUG(printk("DSPOSFX "));
	  len = mod_firmware_load(ZA2FIRMWAREDIR "/dspfx.sim", (char **) &firmware);
	  break;
#endif
	case RESAMPLE_48244 :
	  DEBUG(printk("48244 "));
	  len = mod_firmware_load(ZA2FIRMWAREDIR "/48244.sim", (char **) &firmware);
	  break;
	case RESAMPLE_32244 :
	  DEBUG(printk("32244 "));
	  len = mod_firmware_load(ZA2FIRMWAREDIR "/32244.sim", (char **) &firmware);
	  break;
	case MPEG2 :
	  DEBUG(printk("MPEG2 "));
	  len = mod_firmware_load(ZA2FIRMWAREDIR "/mpeg.sim", (char **) &firmware);
	  break;
	default : 
	  return ZA2_NEED_TO_RELOAD_FILTER;
	}
      if(!firmware)
	{
	  DEBUG(printk("ZA2: Error: Cannot load new filter !\n"));
	  return ZA2_NEED_TO_RELOAD_FILTER;
	}
      DEBUG(printk("Ok\n"));
      if(up_sim(dev,firmware,len,nr))
	{
	  DEBUG(printk("ZA2: Error while uploading dsp-code\n"));
	  dev->filter=INVALID;
	  return ZA2_NEED_TO_RELOAD_FILTER;
	}
      vfree(firmware);
    }
  else return(1);
  return(0);
}

/*** ZaCalcSampleRate ***/
/*fixed point arithmetic: 14 binaries right of the decimal(?) point */
/*for older models of the ZA2 that use (perhaps) the CS4920 we may */
/*just count up to 40 instead of 1024 - this may fix the playback */
/*problems at around 44100 Hz*/
#define ONE (1<<14)
#ifndef CS4920
#define MAXN 1024
#else
#define MAXN 40
#endif
static unsigned int ZaCalcSampleRate(int nsrate)
{
  unsigned short r,bestr=0;
  int z;
  int cm1;
  int div,mlt;
  int nuy,besty=0,nux;
  int bestz=0;
  for(r = 1; r < MAXN; r++)
    {
      nux = ((96000*ONE)/nsrate)*r;
      nuy = nux%ONE;                 /* Nachkommastellen */
      z=nux/ONE;                     /* Vorkommastellen */
      if(nuy > (ONE/2))
	z++;
      if((z>1)&&(r>1)&&(z<4096))
	{
	  nuy = nsrate*ONE - ((96000*ONE)/z)*r;
	  if(nuy<0) nuy*=-1;
	  if((nuy<besty) || (!bestz))
	    {
	      besty = nuy;
	      bestr = r;
	      bestz = z;
	    }
	}
    }
  div = bestz;
  mlt = bestr;
  cm1 = ((div-1)<<10)+mlt-1;
  return(cm1);
}

/* setza */
short setza(za2_status_t *dev)
{
  short  err = 0;                                       /* respond from send command (sync or not) */
  int tmpData = 0;                                      /* temporary variable */
  ZaOpenDSP(dev);
  ZaSendDSPCMD(dev,ZA2_DAC_STATE, 0x9001);              /* disable DAC*/
  ZaSendDSPCMD(dev,ZA2_SAMPLE_RATE,ZaCalcSampleRate(dev->srate));
  if(dev->aeso)
    {
      ZaSendDSPCMD(dev,ZA2_USER_BIT, 0);                /* no user bits in AES */
      tmpData = 0x000300;
      if((dev->srate>=44000) && (dev->srate<44200))
	tmpData += 0x300000;                            /* label as 44.1khz */
      else if(dev->srate<44000)
	tmpData += 0xf00000;                            /* label as 32khz */
      else
	tmpData += 0xc00000;                            /* label as 48khz */
      if(dev->emph)
	tmpData += 0x00f000;                            /* with emphasis */
      ZaSendDSPCMD(dev,ZA2_CBITS_0, tmpData);           /* pro mode */
      ZaSendDSPCMD(dev,ZA2_CBITS_1, 0);
      ZaSendDSPCMD(dev,ZA2_CBITS_2, 0);
      ZaSendDSPCMD(dev,ZA2_CBITS_3, 0);
    }
  else /*SPDIF*/
    {
      ZaSendDSPCMD(dev,ZA2_CBITS_0, 0x003000*(1-dev->scms)+0x00c000*dev->emph);
      ZaSendDSPCMD(dev,ZA2_CBITS_1, 0xc00f00);          /* original DAT source */
      ZaSendDSPCMD(dev,ZA2_CBITS_2, 0x000000);          /* source/channel num */
      if((dev->srate>=44000) && (dev->srate<44200))
	{
	  ZaSendDSPCMD(dev,ZA2_USER_BIT, 1323);
	  ZaSendDSPCMD(dev,ZA2_CBITS_3, 0x000000);      /* label as 44.1khz */
	}
      else if(dev->srate<44000)
	{
	  ZaSendDSPCMD(dev,ZA2_USER_BIT, 1920);
	  ZaSendDSPCMD(dev,ZA2_CBITS_3, 0x000f00);      /* label as 32khz */
	}
      else
	{
	  ZaSendDSPCMD(dev,ZA2_USER_BIT, 1440);
	  ZaSendDSPCMD(dev,ZA2_CBITS_3, 0x000c00);      /* label as 48khz */
	}
    }
  ZaSendDSPCMD(dev,ZA2_SOFT_RESET, 1);                  /* soft reset... */
  ZaSendDSPCMD(dev,ZA2_DO_IT, 0xffffff);                /* do it...needs around 13000 frames*/
  mdelay(220);
  if(dev->currentState == ZA2_CLOSE)
    {
      ZaOpenDSP(dev);
      ZaSendDSPCMD(dev,ZA2_DAC_STATE, 0xb001);          /* enable DAC */  
      err = ZaCloseDSP(dev);
      disable_za2(dev);
    }
  return err;
}

/*** zaSetSamplingRate ***/
short zaSetSamplingRate(za2_status_t *dev,int newsrate)
{
  if( (dev->currentState == ZA2_CLOSE) ||
      (!(dev->currentState&ZA2_OPEN_REC) && !dev->playing) )
    {
      if( (newsrate < 20000) || (newsrate > 60000) )
	return ZA2_VALUE_OUT_OF_BOUND;
      
      dev->srate = newsrate;
      setza(dev);
      return 0;
    }
  if(dev->playing)
    return ZA2_DEVICE_IS_PLAYING;
  return ZA2_DEVICE_NOT_IN_PLAY;
}
