/*  ZA2 Driver for Linux
    Copyright (C) 1998,1999,2000 Peter Wahl

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Send bug reports to:
    Peter Wahl
    uzsa69@uni-bonn.de
*/

#ifndef _ZA2HARD_H
#define _ZA2HARD_H
#include <linux/sched.h>
#ifndef NO_SPINLOCKS
#include <asm/spinlock.h>
#endif
/* -------------------------- */
/*  registers of za2 card     */
/* -------------------------- */

#define ZA2_VOLUME       129
#ifdef DSPOSFX_SUPPORT
/*Volume for dsposfx*/
#define ZA2_ANAVOLUME_L  129
#define ZA2_ANAVOLUME_R  130
#define ZA2_DIGVOLUME_L  131
#define ZA2_DIGVOLUME_R  132
#define ZA2_OUTMODE      133
/*dsposfx end*/
#endif
#define ZA2_DAC_STATE    136
#define ZA2_SAMPLE_RATE  137
#define ZA2_USER_BIT     138
#define ZA2_START_ID     139

#define ZA2_IRQ_RATE     142
#define ZA2_IRQ_COUNT    144
#define ZA2_SYNC_OUTPUT  146   /* output: 0:no output, 1: normal output, 2: R and L swapped, 3: difference R-L */

#define ZA2_SOFT_RESET   157   /* reset if set to 1 */
     
#define ZA2_CBITS_0      224   /* header of cbits stream */
#define ZA2_CBITS_1      225   /* spdif: source: dat, cd, etc. */
#define ZA2_CBITS_2      226   /* spdif: source/channel number */
#define ZA2_CBITS_3      227   /* aes/ebu: sampling rate label */
/*   ...   */
#define ZA2_CBITS_23     247   /* my guess ! */

#define ZA2_DO_IT        255   /* ??? */

/* ------------------------------------------------- */
/*                 error codes                       */
/* ------------------------------------------------- */

#define ZA2_BAD_BIT_FILE              -11

#define ZA2_CANT_CONFIG_ZA2_WITH_ZA1  -13

#define ZA2_DID_NOT_CONFIG_CORRECTLY  -15

#define ZA2_DOWNLOAD_FAILED           -21
#define ZA2_UPSIM_UNKNOW_ERROR        -22

#define ZA2_EXT_SYNC_NOT_VALID        -31
#define ZA2_NEED_TO_RELOAD_FILTER     -32

#define ZA2_NO_DMA_AVAILABLE          -50
#define ZA2_NO_IRQ_AVAILABLE          -51

#define ZA2_UNKNOW_PARAM              -100
#define ZA2_VALUE_OUT_OF_BOUND        -101
#define ZA2_DEVICE_ALREADY_OPEN       -102
#define ZA2_DEVICE_NOT_IN_PLAY        -104
#define ZA2_DEVICE_NOT_IN_REC         -105
#define ZA2_DEVICE_NOT_IN_RECPLAY     -106
#define ZA2_DEVICE_IS_PLAYING         -109
#define ZA2_DEVICE_IS_RECORDING       -110
#define ZA2_NO_INPUT_SIGNAL           -111
#define ZA2_NO_SYNC_RECEIVED          -112
#define ZA2_ACCUMULATED_RX_ERROR      -113
#define ZA2_NO_IRQ_RECEIVED           -115
#define ZA2_FEATURE_NOT_SUPPORTED     -116

#ifndef NO_SPINLOCKS
#define GET_CPU_EXCLUSIVITY spin_lock_irqsave(&dev->za2_lock,flags)

#define RELEASE_CPU_EXCLUSIVITY spin_unlock_irqrestore(&dev->za2_lock,flags)
#else
#define GET_CPU_EXCLUSIVITY do { save_flags(flags); cli(); } while(0)
#define RELEASE_CPU_EXCLUSIVITY restore_flags(flags)
#endif

#define MAX_LOOP 10000

#define ZA2_INTR_TIMEOUT     2000

#define ZA2_WAKEUP_PROCESS wake_up_interruptible(&dev->wq)

#ifndef NO_TIMEOUTS
#define ZA2_BLOCK_PROCESS interruptible_sleep_on_timeout(&dev->wq,ZA2_INTR_TIMEOUT)
#define delay(val) sleep_on_timeout(&dev->wq,val/10)
#else
#define ZA2_BLOCK_PROCESS do {current->timeout = jiffies + ZA2_INTR_TIMEOUT; \
                              interruptible_sleep_on(&dev->wq); } while(0)
#define delay(val) do { current->timeout = jiffies + val/10; sleep_on(&dev->wq);} while(0)
#endif

extern unsigned short __inline__ za2status(za2_status_t *dev);
extern void __inline__ za2command(za2_status_t *dev,short val);
extern void __inline__ disable_za2(za2_status_t *dev);
extern void __inline__ mute_za2(za2_status_t *dev);
void __inline__  za2setmode(za2_status_t *dev);
extern short wait_za2(za2_status_t *dev);
extern void ZaSetEnv(za2_status_t *dev);
extern unsigned int zaGetSamplingRate(za2_status_t *dev);
extern int loadfilter(za2_status_t *dev,int nr);
extern short zaSetSamplingRate(za2_status_t *dev,int newsrate);
extern short setza(za2_status_t *dev);
#ifndef _za2hard_c
extern void ZaOpenDSP(za2_status_t *dev);
extern void ZaSendDSPCMD(za2_status_t *dev,unsigned short cmd, unsigned int data);
extern short ZaCloseDSP(za2_status_t *dev);
extern short ZaNuMode(za2_status_t *dev,short numode);
#endif
#endif

