/*  ZA2 Driver for Linux
    Copyright (C) 1998,1999,2000 Peter Wahl

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Send bug reports to:
    Peter Wahl
    uzsa69@uni-bonn.de
*/

/* HP 20000231  some mods by Heiko Purnhagen  <purnhage@tnt.uni-hannover.de> */

#define __NO_VERSION__
#include "za2.h"
#include <linux/malloc.h>
#include <linux/soundcard.h>
#include <asm/string.h>
#ifndef NO_UACCESS
#include <asm/uaccess.h>
#endif
#include <asm/dma.h>
#include <asm/io.h>

#include "za2hard.h"
#include "za2specials.h"
#include "za2oldkernels.h"

#ifdef DSPOSFX_SUPPORT
void za2setvolume(za2_status_t *dev,int fl)
#else
void za2setvolume(za2_status_t *dev)
#endif
{
  unsigned long flags;
  GET_CPU_EXCLUSIVITY;
  ZaOpenDSP(dev);
#ifdef DSPOSFX_SUPPORT
  if(dev->filter==DSPOSFX)
    {
      if(!fl)
	{
	  ZaSendDSPCMD(dev,ZA2_ANAVOLUME_L,dev->lavol<<8);
	  ZaSendDSPCMD(dev,ZA2_ANAVOLUME_R,dev->ravol<<8);
	}
      else
	{
	  ZaSendDSPCMD(dev,ZA2_DIGVOLUME_L,dev->ldvol<<8);
	  ZaSendDSPCMD(dev,ZA2_DIGVOLUME_R,dev->rdvol<<8);
	}
    }
  else
#endif
    ZaSendDSPCMD(dev,ZA2_VOLUME,dev->lavol<<8);
  ZaCloseDSP(dev);
  RELEASE_CPU_EXCLUSIVITY;
}

void za2startid(za2_status_t *dev)
{
  unsigned long flags;
  GET_CPU_EXCLUSIVITY;
  ZaOpenDSP(dev);
  ZaSendDSPCMD(dev,ZA2_START_ID,300);
  ZaCloseDSP(dev);
  RELEASE_CPU_EXCLUSIVITY;
}

void za2setflags(za2_status_t *dev)
{
  unsigned long flags;
  if(
#ifdef DSPOSFX_SUPPORT
     (dev->filter==DSPOSFX)||
#endif
     (!dev->aeso))
    {
      GET_CPU_EXCLUSIVITY;
      ZaOpenDSP(dev);
      ZaSendDSPCMD(dev,ZA2_CBITS_0,0x003000*(1-dev->scms)+0x00c000*dev->emph);
#ifdef DSPOSFX_SUPPORT
      if(dev->filter==DSPOSFX)
	ZaSendDSPCMD(dev,ZA2_OUTMODE,dev->outmode);
#endif
      ZaCloseDSP(dev);
      RELEASE_CPU_EXCLUSIVITY;
    }
}

void za2setaeso(za2_status_t *dev)
{
  unsigned long flags;
  GET_CPU_EXCLUSIVITY;
  setza(dev);
  /*za2setmode(dev);*/
  RELEASE_CPU_EXCLUSIVITY;
}

/*To be tested: CPU_EXCLUSIVITY for readcbits/readubits:
 * is it worth it ? (->is it more reliable ?)
 * does pose problems for recording/playback
 * is it working ? */
int za2readcbits(za2_status_t *dev,char *bits)
{
  int i=200,b;
  unsigned long flags;
  GET_CPU_EXCLUSIVITY;
  while((za2status(dev)&0x100)&&i)
    {
      wait_za2(dev);
      i--;
    }
  while((!(za2status(dev)&0x100))&&i)
    {
      wait_za2(dev);
      i--;
    }
  if(!i) 
    {
      RELEASE_CPU_EXCLUSIVITY;
      DEBUG(printk("ZA2: readcbits: Timeout !\n"));
      return -EIO;
    }
  for(i=0;i<24;i++,bits++)
    {
      *bits=0;
      for(b=1;b&0xff;b<<=1)
	{
	  if(za2status(dev)&0x40) 
	    *bits|=b;
	  wait_za2(dev);
	}
    }
  RELEASE_CPU_EXCLUSIVITY;
  return 0;
}

int __inline__ wait_sample(za2_status_t *dev,int last)
{
#ifdef USE_DELAY
  int err=40;
#else
  int err=100;
#endif
  while(((za2status(dev)&8)==(last&8)) && --err) 
#ifdef USE_DELAY
    udelay(1)
#endif
      ;
  if(err)
    return za2status(dev);
  else 
    {
      DEBUG(printk("ZA2: wait_sample: Timeout !\n"));
      return -1;
    }
}

int za2readubits(za2_status_t *dev,char *bits)
{
  unsigned long flags;
  int c=0,b,err=1200,last;
  unsigned long int crc;
  GET_CPU_EXCLUSIVITY;
  last=za2status(dev);
  while((c<16)&&err)
    {
      last=wait_sample(dev,last);
      if(last&0x80) c=0;
      else c++;
      err--;
    }
  if(!err) 
    {
      RELEASE_CPU_EXCLUSIVITY;
      DEBUG(printk("ZA2: readubits: Can't find start of UBITS !\n"));
      return -EIO;
    }
  for(c=0;c<98;c++)
    {
      bits[c]=1; err=100;
      last=za2status(dev);
      while((!(last&0x80))&&err)
	{
	  last=wait_sample(dev,last);
	  err--;
	}
      if(!err) 
	{
	  RELEASE_CPU_EXCLUSIVITY;
	  DEBUG(printk("ZA2: readubits: Timeout !\n"));
	  return -EIO;
	}
      for(b=2;b&0xff;b<<=1)
	{ 
	  last=wait_sample(dev,last);
	  if(last&0x80) bits[c]|=b;
	}
    }
  RELEASE_CPU_EXCLUSIVITY;
  for(c=0,crc=0;c<80;c++)
    {
      crc = crc<<1;
      if (bits[c]&2)
	crc++;
      if (crc&0x10000)
	crc^=0x11021;
    }
  for(c=80;c<96;c++)
    {
      crc = crc<<1;
      if (!(bits[c]&2))
	crc++;
      if (crc&0x10000)
	crc^=0x11021;
    }
  if(crc) 
    {
      DEBUG(printk("ZA2: readubits: CRC error in UBIT-stream.\n"));
      return -EIO;
    }
  else return 0;
}

int za2_mixer(za2_status_t *dev,unsigned int cmd, unsigned long *param)
{
  unsigned int recsrc;
  switch(cmd) 
    {
    case SOUND_MIXER_WRITE_VOLUME:
#ifdef DSPOSFX_SUPPORT
      if(dev->mixer_devs&SOUND_MASK_VOLUME)
	{
#endif
	  if((*param&0xff)>100) return -EINVAL;
	  dev->lavol=((*param&0xff)<<14)/100; /* scale volume from 0..100 to 0..16384 */
#ifdef DSPOSFX_SUPPORT
	  if(dev->mixer_caps&SOUND_MASK_VOLUME) 
	    {
	      if((*param&0xff00)>(100<<8)) return -EINVAL;
	      dev->ravol=((*param&0xff00)<<6)/100;
	    }
	  za2setvolume(dev,0);
#else
	  za2setvolume(dev);
#endif
#ifdef DSPOSFX_SUPPORT
	}
      else return -EINVAL;
#endif
      break;
#ifdef DSPOSFX_SUPPORT
    case SOUND_MIXER_WRITE_LINE:
      if(dev->mixer_devs&SOUND_MASK_LINE)
	{
	  if((*param&0xff)>100) return -EINVAL;
	  dev->ldvol=((*param&0xff)<<14)/100; /* scale volume from 0..100 to 0..16384 */
	  if(dev->mixer_caps&SOUND_MASK_LINE) 
	    {
	      if((*param&0xff00)>(100<<8)) return -EINVAL;
	      dev->rdvol=((*param&0xff00)<<6)/100;
	    }
	  za2setvolume(dev,1);
	}
      else return -EINVAL;
      break;
#endif
    case SOUND_MIXER_READ_VOLUME:
#ifdef DSPOSFX_SUPPORT
      if(dev->mixer_devs&SOUND_MASK_VOLUME)
	{
#endif
	  *param=(dev->lavol*100)>>14;
#ifdef DSPOSFX_SUPPORT
	  if(dev->mixer_caps&SOUND_MASK_VOLUME)
	    *param|=((dev->ravol*100)>>6)&0xff00;
	}
      else return -EINVAL;
#endif
      break;
#ifdef DSPOSFX_SUPPORT
    case SOUND_MIXER_READ_LINE:
      if(dev->mixer_devs&SOUND_MASK_LINE)
	{
	  *param=(dev->ldvol*100)>>14;
	  if(dev->mixer_caps&SOUND_MASK_LINE)
	    *param|=((dev->rdvol*100)>>6)&0xff00;
	}
      else return -EINVAL;
      break;
#endif
    case SOUND_MIXER_WRITE_RECSRC:
      /* patch to disable current recsrc in requested recsrc   HP 20000119 */
      switch(dev->digin)
	{
	case 0 : recsrc=SOUND_MASK_DIGITAL1; /* TOSLINK */
	  break;
	case 1 : recsrc=SOUND_MASK_DIGITAL2; /* COAX */
	  break;
	case 2 : recsrc=SOUND_MASK_DIGITAL3; /* AES/EBU */
	  break;
	default: return -EINVAL;
	}
      switch(*param & ~recsrc)
	{
	case SOUND_MASK_DIGITAL1 : dev->digin=0; /* TOSLINK */
	  ZaSetEnv(dev);
	  za2setmode(dev);
	  break;
	case SOUND_MASK_DIGITAL2 : dev->digin=1; /* COAX */
	  ZaSetEnv(dev);
	  za2setmode(dev);
	  break;
	case SOUND_MASK_DIGITAL3 : dev->digin=2; /* AES/EBU */
	  ZaSetEnv(dev);
	  za2setmode(dev);
	  break;
	case 0: if(*param) break; /* requested current recsrc */
	default: return -EINVAL;
	}
      break;
    case SOUND_MIXER_READ_DEVMASK:
      *param=SOUND_MASK_DIGITAL1|SOUND_MASK_DIGITAL2|SOUND_MASK_DIGITAL3
#ifdef DSPOSFX_SUPPORT
	|dev->mixer_devs
#endif
	;
      break;
    case SOUND_MIXER_READ_STEREODEVS:
      *param=SOUND_MASK_DIGITAL1|SOUND_MASK_DIGITAL2|SOUND_MASK_DIGITAL3
#ifdef DSPOSFX_SUPPORT
	|dev->mixer_caps
#endif
	;
      break;
    case SOUND_MIXER_READ_RECMASK:
      *param=SOUND_MASK_DIGITAL1|SOUND_MASK_DIGITAL2|SOUND_MASK_DIGITAL3;
      break;
    case SOUND_MIXER_READ_RECSRC:
      switch(dev->digin)
	{
	case 0 : *param=SOUND_MASK_DIGITAL1; /* TOSLINK */
	  break;
	case 1 : *param=SOUND_MASK_DIGITAL2; /* COAX */
	  break;
	case 2 : *param=SOUND_MASK_DIGITAL3; /* AES/EBU */
	  break;
	default: return -EINVAL;
	}
      break;
    case SOUND_MIXER_READ_CAPS:
      *param=SOUND_CAP_EXCL_INPUT;
      break;
    case SOUND_MIXER_ZA2SETFLAGS:                 /*set SCMS, EMPH, AESO*/
      dev->scms=(*param&SOUND_MIXER_ZA2SCMS)?1:0;
      dev->emph=(*param&SOUND_MIXER_ZA2EMPH)?1:0;
#ifdef DSPOSFX_SUPPORT
      dev->outmode=((*param&SOUND_MIXER_ZA2SWAP)?8:0)+
	((*param&SOUND_MIXER_ZA2DIFF)?4:0)+
	((*param&SOUND_MIXER_ZA2ADIRECT)?1:0)+
	((*param&SOUND_MIXER_ZA2DDIRECT)?2:0);
#endif
      if((dev->aeso<<2)!=(*param&SOUND_MIXER_ZA2AESO))
	{
	  dev->aeso=(*param&SOUND_MIXER_ZA2AESO)?1:0;
	  if(!dev->playing) za2setaeso(dev); /*<-- the if(!...) can be removed => glitches !*/
	}
      za2setflags(dev);
      break;
    case SOUND_MIXER_ZA2GETFLAGS:                   /*get SCMS, EMPH, AESO*/
      *param=((dev->emph)?SOUND_MIXER_ZA2EMPH:0)+
	((dev->scms)?SOUND_MIXER_ZA2SCMS:0)+
	((dev->aeso)?SOUND_MIXER_ZA2AESO:0)
#ifdef DSPOSFX_SUPPORT
	+((dev->outmode&8)?SOUND_MIXER_ZA2SWAP:0)+
	((dev->outmode&4)?SOUND_MIXER_ZA2DIFF:0)+
	((dev->outmode&1)?SOUND_MIXER_ZA2ADIRECT:0)+
	((dev->outmode&2)?SOUND_MIXER_ZA2DDIRECT:0)
#endif
	;
      break;
    case SOUND_MIXER_ZA2STARTID:
      za2startid(dev);
      break;
    default:
      return -EINVAL;
    }
  return 0;
}

static int mixer_ioctl(struct inode * aInode, struct file * aFile,
		       unsigned int cmd, unsigned long arg)
{
  int err;
  long param = 0;
  char *bits;
  za2_status_t *dev=za2_st;
  /* extract the parameter */
  if(((cmd>>8)&0xff)=='M')
     if(cmd==SOUND_MIXER_ZA2GETUBITS)
       {
	 bits=(char *)kmalloc(ZA2UBITS+1,GFP_KERNEL);
	 if(!bits) return -ENOMEM;
	 err=za2readubits(dev,bits);
	 /*if(!err)*/ copy_to_user((void *) arg,bits,ZA2UBITS);
	 kfree(bits);
	 return err;
       }
     else if(cmd==SOUND_MIXER_ZA2GETCBITS)
       {
	 bits=(char *)kmalloc(ZA2CBITS+1,GFP_KERNEL);
	 if(!bits) return -ENOMEM;
	 err=za2readcbits(dev,bits);
	 copy_to_user((void *) arg,bits,ZA2CBITS);
	 kfree(bits);
	 return err;
       }
     else
       {
	 if((_SIOC_DIR(cmd)&_SIOC_WRITE)&&(_SIOC_SIZE(cmd)<=sizeof(long)))
	   {
	     if((err=verify_area(VERIFY_READ, (void *) arg, sizeof(long))))
	       return err;
#ifndef OLD_GET_USER
	     get_user(param,(long *) arg);
#else
	     param=get_user((long *) arg);
#endif
	   }
	 /* do the command */
	 if((err=za2_mixer(dev,cmd,&param))) return err;
	 /* put back the parameter */
	 if((_SIOC_DIR(cmd)&_SIOC_READ)&&(_SIOC_SIZE(cmd)<=sizeof(long))) 
	   {
	     if((err=verify_area(VERIFY_WRITE, (void *) arg, sizeof(long))))
	       return err;
	     put_user(param, (long *) arg);
	   }
	 return 0;  /* all other settings are correct, but not handled */
       }
  else return -EINVAL;
}

#ifndef NO_SOUND_H
static int mixer_open(struct inode * aInode, struct file * aFile)
{
#ifdef USE_MODCOUNT
  MOD_INC_USE_COUNT;
#endif
  return 0;
}

static int mixer_release(struct inode * aInode, struct file * aFile)
{
#ifdef USE_MODCOUNT
  MOD_DEC_USE_COUNT;
#endif
  return 0;
}

static struct file_operations za2mixer_fops = {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
  THIS_MODULE,
#endif
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  mixer_ioctl,
  NULL,
  mixer_open,
#ifndef OLD_FILEOPS
  NULL,
#endif
  mixer_release,
  NULL,
  NULL,
  NULL,
  NULL
#ifndef OLD_FILEOPS
, NULL
#endif
};
#endif
